import os
import sys
from pathlib import Path

import click

from .types import Store, YopError, YubiKey
from .utils import combine_credentials, find_actionable_credentials, generate_table


@click.command
@click.argument(
    "store",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True, path_type=Path),
)
@click.option(
    "--really",
    is_flag=True,
    show_default=True,
    default=False,
    help="Really execute the sync (aka the opposite of dry-run)",
)
@click.option(
    "--delete",
    is_flag=True,
    show_default=True,
    default=False,
    help="Delete YubiKey credentials not present in pass",
)
def cli(store: Path, really: bool, delete: bool) -> None:
    """
    Synchronize OTP credentials in the currently plugged in YubiKey (one YubiKey at a time is supported), with those
    stored in pass (https://www.passwordstore.org/), at the filesystem path specified by STORE.

    The OTP credentials that are created on the YubiKey, follow the naming scheme of the `ykman oath` command line
    tool: `<issuer>:<name>` (e.g., `paypal.com:me@example.com`).

    In pass, the filename is assumed to be the issuer, the first line of the file - the secret. If a line in the file
    starts with `user:` or `username:`, the value after `: ` is assumed to be the credential name; otherwise, the entire
    second line is assumed to be the name.

    Dry-run is the default, use '--really' to disable. Credentials, not found in pass, are not deleted from the YubiKey.
    Use `--delete` to force.

    Parsing pass credentials is atomic: the operation is aborted if parsing fails, or if pass contains more than 32
    credentials (YubiKey OTP app limitation).
    """
    os.environ["PASSWORD_STORE_DIR"] = str(store)

    try:
        credential_store = Store(store)
    except YopError as e:
        click.echo(click.style(e, fg="red"))
        sys.exit(1)

    try:
        yubikey = YubiKey.detect()
    except YopError as e:
        click.echo(click.style(e, fg="red"))
        sys.exit(1)

    click.echo(f"Reading credentials in {store} and in {yubikey.info}. This may take a moment...")
    click.echo()

    try:
        store_credentials = credential_store.collect_credentials()
    except YopError as e:
        click.echo(click.style(e, fg="red"))
        sys.exit(1)

    num_store_credentials = len(store_credentials)

    if num_store_credentials > 32:
        msg = f"YubiKey OTP app allows up to 32 credentials, but {num_store_credentials} credentials found in {store}"
        click.echo(click.style(msg, fg="red"))
        sys.exit(1)

    yubikey_credentials = yubikey.collect_credentials()

    combined_credentials = combine_credentials(store_credentials, yubikey_credentials)

    click.echo(generate_table(combined_credentials))

    to_add, to_delete = find_actionable_credentials(combined_credentials)

    if not any((to_add, to_delete)):
        click.echo()
        click.echo(click.style(f"Credentials on YubiKey are in sync with {store}", fg="green"))
        sys.exit()

    if really:
        click.echo()

        with yubikey.get_session() as session:
            if to_add:
                click.echo("Adding credentials to YubiKey")
                for cred in to_add:
                    click.echo(cred)
                    cred.write_to_yubikey(session)

            if to_delete:
                # Separate the two with an empty line
                if to_add:
                    click.echo()

                if delete:
                    click.echo("Deleting credentials from YubiKey")
                    for cred in to_delete:
                        click.echo(cred)
                        cred.delete_from_yubikey(session)
                else:
                    click.echo(click.style("Skipping deletion. Use '--delete' to force", fg="yellow"))

    else:
        click.echo()
        click.echo(click.style("In dry-run mode. Use '--really' to disable", fg="yellow"))


# https://stackoverflow.com/a/45881691
def safe_cli() -> None:
    try:
        cli()
    except Exception as e:
        click.echo(click.style(e, fg="red"))



from typing import Dict, List, Tuple

import click
from tabulate import tabulate

from .types import Credential


def combine_credentials(store_credentials: List[Credential], yubikey_credentials: List[Credential]) -> Dict:
    credentials = {cred: {"store_path": cred.store_path, "exists_in_yubikey": False} for cred in store_credentials}

    for cred in yubikey_credentials:
        if cred in credentials:
            credentials[cred]["exists_in_yubikey"] = True
        else:
            credentials[cred] = {"store_path": None, "exists_in_yubikey": True}

    return credentials


def generate_table(credential_data: Dict) -> str:
    table_header = ["Credential", "Store", "YubiKey"]
    table = []

    for cred_as_str, meta in credential_data.items():
        table.append(
            [
                cred_as_str,
                meta["store_path"] if meta["store_path"] else click.style("n/a", fg="red"),
                "y" if meta["exists_in_yubikey"] else click.style("n", fg="red"),
            ]
        )

    return tabulate(table, headers=table_header)


def find_actionable_credentials(credentials: Dict) -> Tuple[List[Credential], List[Credential]]:
    to_add = [cred for cred, meta in credentials.items() if not meta["exists_in_yubikey"]]
    to_delete = [cred for cred, meta in credentials.items() if meta["exists_in_yubikey"] and meta["store_path"] is None]
    return to_add, to_delete



import subprocess
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterator, List, Optional, Tuple

from ykman.base import YkmanDevice
from ykman.device import list_all_devices
from yubikit.core.smartcard import SmartCardConnection
from yubikit.oath import HASH_ALGORITHM, OATH_TYPE
from yubikit.oath import Credential as YubikeyCredential
from yubikit.oath import CredentialData, OathSession, parse_b32_key
from yubikit.support import get_name, read_info


class CredentialParseError(Exception):
    pass


class YopError(Exception):
    pass


@dataclass
class Credential:
    issuer: str
    name: str
    secret: Optional[str] = None
    store_path: Optional[Path] = None

    def __hash__(self) -> int:
        return hash((self.issuer, self.name))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Credential):
            return NotImplemented

        return self.issuer == other.issuer and self.name == other.name

    def __repr__(self) -> str:
        secret = None if self.secret is None else "<REDACTED>"
        return f"Credential(issuer='{self.issuer}', name='{self.name}', secret={secret})"

    def __str__(self) -> str:
        return self.id

    @property
    def id(self) -> str:
        return f"{self.issuer}:{self.name}"

    @staticmethod
    def from_path(path: Path) -> "Credential":
        resp = subprocess.run(["pass", "show", path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        resp.check_returncode()

        name, secret = Credential.parse_credential_data(resp.stdout.decode())
        return Credential(issuer=path.name, name=name, secret=secret, store_path=path)

    @staticmethod
    def from_yubikey_credential(yubikey_credential: YubikeyCredential) -> "Credential":
        if yubikey_credential.issuer is None:
            raise YopError("Credential issuer is not set")

        return Credential(issuer=yubikey_credential.issuer, name=yubikey_credential.name)

    @staticmethod
    def parse_credential_data(data: str) -> Tuple[str, str]:
        lines = [line.strip() for line in data.splitlines() if not line.isspace()]
        if not lines:
            raise CredentialParseError("File is empty")

        secret = lines.pop(0)
        try:
            parse_b32_key(secret)
        except Exception:
            raise CredentialParseError("Value found is not an OTP secret")

        if not lines:
            raise CredentialParseError("File does not contain a username")

        for line in lines:
            if line.lower().startswith(("name:", "username:")):
                name = line.split(":")[-1].strip()
        else:
            name = lines[0]

        return name, secret

    def write_to_yubikey(self, session: OathSession) -> None:
        session.put_credential(self.as_yubikey_credential_data())

    def delete_from_yubikey(self, session: OathSession) -> None:
        session.delete_credential(self.id.encode())

    def as_yubikey_credential_data(self) -> CredentialData:
        if self.secret is None:
            raise YopError("Credential secret is not set")

        return CredentialData(
            name=self.id,
            oath_type=OATH_TYPE.TOTP,
            hash_algorithm=HASH_ALGORITHM.SHA1,
            secret=self.secret.encode(),
        )


@dataclass
class Store:
    path: Path

    def __post_init__(self) -> None:
        if not (self.path / ".gpg-id").exists():
            raise YopError(f"{self.path} is not a pass directory")

    def collect_credentials(self) -> List[Credential]:
        credentials = []

        for gpg_file in Path(self.path).rglob("*.gpg"):
            pass_path = gpg_file.relative_to(self.path).with_suffix("")

            try:
                credentials.append(Credential.from_path(pass_path))
            except subprocess.CalledProcessError as e:
                err = ". ".join([m.decode().replace("\n", ". ") for m in (e.stdout, e.stderr) if m])
                raise YopError(f"Failed to read credential {pass_path}: {err}")
            except CredentialParseError as e:
                raise YopError(f"Failed to parse credential {pass_path}: {e}")

        return credentials


@dataclass
class YubiKey:
    device: YkmanDevice
    info: str = field(init=False)

    def __post_init__(self) -> None:
        with self.get_connection() as conn:
            info = read_info(conn, self.device.pid)
            key_type = None if self.device.pid is None else self.device.pid.yubikey_type
            self.info = f"{get_name(info, key_type)} (serial number {info.serial})"

    @staticmethod
    def detect() -> "YubiKey":
        connected_yubikeys = list_all_devices()

        if not connected_yubikeys:
            raise YopError("No YubiKeys found")

        if len(connected_yubikeys) > 1:
            raise YopError("Multiple YubiKeys found")

        return YubiKey(device=connected_yubikeys[0][0])

    @contextmanager
    def get_connection(self) -> Iterator[SmartCardConnection]:
        yield self.device.open_connection(SmartCardConnection)

    @contextmanager
    def get_session(self) -> Iterator[OathSession]:
        with self.get_connection() as conn:
            yield OathSession(conn)

    def collect_credentials(self) -> List[Credential]:
        with self.get_session() as session:
            return [Credential.from_yubikey_credential(c) for c in session.list_credentials()]
